
import os
import re


def collect_files(seeds, extensions=None):
    '''
    Collect files from given seeds: files or folders to scan through recursively.

    @param seeds list of root folders or files
    @param extensions optional list of file extensions (lowercase) to filter files
    '''
    files = set()

    # Local function to check extensions (or accept everything)
    if extensions is not None:
        check_extension = lambda path: os.path.splitext(path)[1].lower() in extensions
    else:
        check_extension = lambda path: True

    for seed in seeds:
        if os.path.isfile(seed) and check_extension(seed):
            files.add(seed)
        elif os.path.isdir(seed):
            for (dirpath, dirnames, filenames) in os.walk(seed):
                for filename in filenames:
                    path = os.path.join(dirpath, filename)
                    if check_extension(path):
                        files.add(path)
    return files



def file_get_contents(file_name):
    with open(file_name) as f:
        contents = f.read()
    return contents





def get_occuring_words(words, content):
    '''
    Return the subset of given words that occur in content.
    '''
    found = set()
    for word in words:
        if re.search(r'\b%s\b' % word, content):
            found.add(word)
    return found





import logging

import lxml
import lxml.etree
import lxml.cssselect
import cssselect


# Global logging object
_log = logging.getLogger('cssdeadwood.dom_match')




class CssDeadwoodHtmlTranslator(cssselect.HTMLTranslator):
    '''
    Simple extension of cssselect.HTMLTranslator to make sure that
    pseudo classes like :hover, :focus, :visited, etc always match,
    which is what we want in a dead code detection app.
    '''

    def pseudo_always_matches(self, xpath):
        """Common implementation for pseudo-classes that alwyas match."""
        return xpath

    xpath_link_pseudo = pseudo_always_matches
    xpath_visited_pseudo = pseudo_always_matches
    xpath_hover_pseudo = pseudo_always_matches
    xpath_active_pseudo = pseudo_always_matches
    xpath_focus_pseudo = pseudo_always_matches
    xpath_target_pseudo = pseudo_always_matches
    xpath_enabled_pseudo = pseudo_always_matches
    xpath_disabled_pseudo = pseudo_always_matches
    xpath_checked_pseudo = pseudo_always_matches


def match_selectors_against_html_root_element(selectors, html_element):
    '''
    Find the selectors that match with the DOM from the given HTML.

    @param selectors set of CSS selectors (strings)
    @param html_element lxml.etree.Element object

    @return set of found selectors
    '''
    found_selectors = set()
    css_to_xpath_translator = CssDeadwoodHtmlTranslator()
    for selector_str in selectors:
        try:
            # Instead of just calling css_to_xpath(selector_str),
            # we first convert the css selector string to a cssselect.Selector instance
            # to pass to selector_to_xpath(), so we can properly ignore pseudo elements.
            # Note that cssselect.parse() always returns a list, so we do a for loop.
            for selector in cssselect.parse(selector_str):
                selector.pseudo_element = None
                xpath_expr = css_to_xpath_translator.selector_to_xpath(selector)
                if len(html_element.xpath(xpath_expr)) > 0:
                    found_selectors.add(selector_str)
        except Exception:
            global _log
            _log.exception('lxml css select failed on selector %r' % selector_str)
    return found_selectors


def match_selectors_against_html_string(selectors, html_string):
    '''
    Find the selectors that match with the DOM from the given HTML.

    @param selectors set of CSS selectors (strings)
    @param html_string html string

    @return set of found selectors
    '''
    parser = lxml.etree.HTMLParser()
    html_element = lxml.etree.fromstring(html_string, parser=parser)
    return match_selectors_against_html_root_element(selectors, html_element)


def match_selectors_against_html_resource(selectors, html_resource):
    '''
    Find the selectors that match with the DOM from the given HTML.

    @param selectors set of CSS selectors (strings)
    @param html_resource HTML file path/url or file(-like) object.

    @return set of found selectors
    '''
    parser = lxml.etree.HTMLParser()
    html_element = lxml.etree.parse(html_resource, parser=parser).getroot()
    return match_selectors_against_html_root_element(selectors, html_element)





import os
import sys
import tempfile
import unittest


from cssdeadwood.dom_match import match_selectors_against_html_string


class CssMatchTest(unittest.TestCase):


    def testSimple(self):
        html = '<html><head></head><body><p>hello world</p></body></html>'

        selectors = set(['p'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p']))

        selectors = set(['div'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set([]))

        selectors = set(['p', 'div'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p']))


    def testPseudoClasses(self):
        html = '<html><head></head><body><p>hello <a href="/world">world</a></p></body></html>'

        selectors = set(['a', 'p:hover', 'h4:hover'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['a', 'p:hover']))

        selectors = set(['a:focus', 'a:visited'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['a:visited', 'a:focus']))


    def testPseudoChildSelectors(self):
        html = '<html><head></head><body><ol><li>one</li></ol><ul></ul></body></html>'

        selectors = set(['ol li:first-child'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['ol li:first-child']))

        selectors = set(['ul li:first-child'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set())


    def testDirectChilds(self):
        html = '<html><head></head><body><p>hello <a href="/world">world</a></p></body></html>'

        selectors = set(['p > a', 'p a'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p > a', 'p a']))

        selectors = set(['p>a', 'p a'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p>a', 'p a']))


    def testPseudoElements(self):
        html = '<html><head></head><body><p>hello <a href="/world">world</a></p></body></html>'

        selectors = set(['p:before'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p:before']))

        selectors = set(['p:after'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p:after']))

        selectors = set(['p:first-line'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p:first-line']))

        selectors = set(['p:first-letter'])
        result = match_selectors_against_html_string(selectors, html)
        self.assertEqual(result, set(['p:first-letter']))






#!/usr/bin/env python


import sys
import logging
import optparse
import json


from cssdeadwood.utils import collect_files, file_get_contents, get_occuring_words
from cssdeadwood.css_extract import extract_css_selectors, extract_ids_and_classes_from_selectors
from cssdeadwood.dom_match import match_selectors_against_html_resource


# TODO: instead of used vs not used, provide histogram analysis to have better view on hot vs not hot
# TODO: source id/class based elimination: only do search on strings, not logic
# TODO: operation mode to only provide HTML file(s) (and analyse referenced CSS files)
# TODO: operation mode to only provide url(s) (and analyse referenced CSS files)
# TODO: HTML format reporting


_log = logging.getLogger('cssdeadwood')





class CssDeadwoodApp(object):
    '''
    CSS Deadwood main() function,
    written OOP-style and chopped up in small methods, for better testability.
    '''


    def _eliminate_selectors_from_dom_matching(self, selectors, html_files):
        # The results struct for tracking intermediate data
        results = {}

        # Start with flagging all selectors as "unused"
        unused_selectors = selectors.copy()
        for html_file in html_files:
            original_total = len(unused_selectors)
            _log.debug('DOM matching %d CSS selectors with DOM from %r' % (original_total, html_file))
            found_selectors = match_selectors_against_html_resource(unused_selectors, html_file)
            unused_selectors.difference_update(found_selectors)
            _log.info('DOM matching %d CSS selectors: %d matches, %d unmatched with DOM from %r' % (original_total, len(found_selectors), len(unused_selectors), html_file))

        # Return result
        results['unused_selectors'] = sorted(unused_selectors)

        return unused_selectors, results

    def _eliminate_selectors_from_idclass_grepping(self, selectors, src_files):
        '''
        Eliminate selectors by searching for mentioned ids and classes in the given source files.
        '''
        # The results struct for tracking intermediate data
        results = {}

        # Extract ids and classes.
        ids, classes, origins = extract_ids_and_classes_from_selectors(selectors)
        results['ids'] = ids
        results['classes'] = classes
        _log.info('Id/class extraction from %d CSS selectors for source code matching: extracted %d ids, %d classes.' % (len(selectors), len(ids), len(classes)))
        _log.debug('Extracted ids: %r' % ids)
        _log.debug('Extracted classes: %r' % classes)

        # Determine unfindable ids and classes.
        findable_ids = set()
        findable_classes = set()
        unfindable_ids = ids.copy()
        unfindable_classes = classes.copy()
        # Scan through the source files for the remaining ids and classes.
        for src_file in src_files:
            content = file_get_contents(src_file)
            _log.debug('Searching for %d remaining unfindable ids in %s' % (len(unfindable_ids), src_file))
            findable_ids.update(get_occuring_words(unfindable_ids, content))
            unfindable_ids.difference_update(findable_ids)
            _log.debug('Searching for %d remaining unfindable classes in %s' % (len(unfindable_classes), src_file))
            findable_classes.update(get_occuring_words(unfindable_classes, content))
            unfindable_classes.difference_update(findable_classes)
        results['unfindable_ids'] = unfindable_ids
        results['unfindable_classes'] = unfindable_classes

        # Eliminate selectors with findable ids/classes
        used_idclass_selectors = set()
        for id in findable_ids:
            used_idclass_selectors.update(origins['#' + id])
        for classs in findable_classes:
            used_idclass_selectors.update(origins['.' + classs])
        unused_selectors = selectors.difference(used_idclass_selectors)
        _log.info('Id/class based elimination from {total:d} CSS selectors with {src:d} source files: {used:d} possibly used, {unused:d} unused.'.format(
            total=len(selectors),
            src=len(src_files),
            used=len(used_idclass_selectors),
            unused=len(unused_selectors)
        ))

        # return result
        results['unused_selectors'] = sorted(unused_selectors)

        return unused_selectors, results


    def main(self, argv=sys.argv):

        # Parse command line
        option_parser = optparse.OptionParser(usage='%prog [options] [cssfiles] [htmlfiles] [srcsfiles]')

        default_src_extensions = '.php,.py,.rb,.js'
        option_parser.add_option("-e", "--srcext", metavar='EXT',
                      action="store", dest="src_extensions", default=default_src_extensions,
                      help="Define the source file extensions (comma separated) to filter on when recursively scanning source folders. Default: '%s'." % default_src_extensions)

        option_parser.add_option("--htmlexport", metavar='FILE',
                      action="store", dest="html_export", default=None,
                      help="Export result to a HTML report (requires jinja2 library).")
        option_parser.add_option("--jsonexport", metavar='FILE',
                      action="store", dest="json_export", default=None,
                      help="Export analysis results in JSON format.")

        option_parser.add_option("-v", "--verbose",
                      action="store_const", dest="loglevel", const=logging.DEBUG, default=logging.INFO,
                      help="Be more verbose")

        options, args = option_parser.parse_args(args=argv[1:])

        # Set up logging
        logging.basicConfig(level=options.loglevel)
        logging.getLogger('CssDeadwood.bs4').setLevel(logging.ERROR)

        # Get CSS, HTML and other source files form given arguments.
        css_files = collect_files(args, extensions=['.css'])
        html_files = collect_files(args, extensions=['.html'])
        if len(options.src_extensions.strip()) > 0:
            src_extensions = options.src_extensions.split(',')
        else:
            src_extensions = []
        src_files = collect_files(args, extensions=src_extensions)
        _log.info('Working with %d CSS files.' % len(css_files))
        _log.debug('CSS files: %r.' % css_files)
        _log.info('Working with %d HTML files.' % len(html_files))
        _log.debug('HTML files: %r.' % html_files)
        _log.info('Working with %d source files.' % len(src_files))
        _log.debug('Source files: %r.' % src_files)

        # Result object where we will store all analysis data, to be used in reporting/exporting.
        results = {}

        for css_file in css_files:
            _log.info('Analysing CSS selectors from %r' % css_file)
            results[css_file] = {}

            # Extract selectore from CSS source
            selectors = extract_css_selectors(css_file)
            results[css_file]['selectors'] = selectors
            _log.info('Extracted %d CSS selectors from %r.' % (len(selectors), css_file))
            _log.debug('Extracted selectors: %r' % selectors)

            # Start with flagging all selectors as "unused"
            unused_selectors = selectors.copy()

            # Eliminate selectors that match with the DOM trees from the HTML files.
            if html_files:
                unused_selectors, data = self._eliminate_selectors_from_dom_matching(unused_selectors, html_files)
                results[css_file]['dom_matching'] = data

            # Extract ids and classes and scan other source files for these.
            if src_files:
                unused_selectors, data = self._eliminate_selectors_from_idclass_grepping(unused_selectors, src_files)
                results[css_file]['idclass_elimination'] = data

            results[css_file]['unused_selectors'] = sorted(unused_selectors)

        # Report
        for css_file, data in results.items():
            print (css_file + ' ').ljust(80, '-')
            total_count = len(data['selectors'])
            if total_count == 0:
                print 'No selectors'
                continue
            unused_count = len(data['unused_selectors'])
            perc = 100.0 * unused_count / total_count
            print 'Could not determine usage of the following %d CSS selectors (from %d in total: %.2f%%):' % (unused_count, total_count, perc)
            print '\n'.join(data['unused_selectors'])

        # TODO: HTML report

        # JSON report
        if options.json_export:
            logging.info('Writing JSON report: %s' % options.json_export)
            with open(options.json_export, 'w') as f:
                json.dump(results, f, indent=1, default=list)



def main():
    CssDeadwoodApp().main(sys.argv)





import re
import collections

from cssdeadwood.utils import file_get_contents



def extract_css_selectors(css_file):
    '''
    Extract CSS selectors from a given CSS file.

    @param css_file CSS file path

    @return set of CSS selectors
    '''
    selectors = set()
    css = file_get_contents(css_file)
    # Precompiled regex to clean up/normalize whitespace
    whitespace_regex = re.compile(r'\s+', flags=re.DOTALL)
    # Remove comments.
    css = re.compile(r'/\*.*?\*/', flags=re.DOTALL).sub('', css)
    # Collect the selectors/selector groups in front of { ... } declaration blocks
    for match in re.compile(r'\s*([^{}]*)\s*{', flags=re.DOTALL).finditer(css):
        selector_part = match.group(1).strip()
        # Ignore at-rules
        if selector_part.startswith('@'):
            continue
        # Split on comma to get selectors.
        for selector_candidate in selector_part.split(','):
            # Clean up/normalize whitespace
            selector_candidate = whitespace_regex.sub(' ', selector_candidate.strip())
            # Store
            selectors.add(selector_candidate)

    return selectors





# Some precompiled regexes to extract ids and clasess from CSS selectors.
REGEX_ID = re.compile(r'\#([a-zA-Z0-9]+)')
REGEX_CLASS = re.compile(r'\.([a-zA-Z0-9]+)')


def extract_ids_and_classes_from_selectors(selectors):
    '''
    Extract ids and classes used in the given CSS selectors.

    @return (ids, classes, origins) with:
        ids: set of extracted ids
        classes: set of extracted classes
        origins: mapping of ids (format '#x') and classes (format '.x')
            to list of selectors they were found in.
    '''
    ids = set()
    classes = set()
    origins = collections.defaultdict(lambda: [])
    for selector in selectors:
        for id in REGEX_ID.findall(selector):
            ids.add(id)
            origins['#' + id].append(selector)
        for classs in REGEX_CLASS.findall(selector):
            classes.add(classs)
            origins['.' + classs].append(selector)
    return (ids, classes, origins)






try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='CssDeadwood',
    version='0.2.5',
    author='Stefaan Lippens',
    author_email='soxofaan@gmail.com',
    url='https://github.com/soxofaan/CssDeadwood/',
    packages=['cssdeadwood', 'cssdeadwood.test'],
    entry_points={
        'console_scripts': [
            'cssdeadwood = cssdeadwood.app:main',
        ],
    },
    license='MIT',
    keywords="css deadcode",
    description='Tool to search CSS files for unused CSS selectors.',
    long_description=open('README.rst').read(),
    install_requires=[
        "lxml >= 3.1.1",
        "cssselect >= 0.8",
    ],
    test_suite='cssdeadwood.test'
)




import os
import sys
import tempfile
import unittest



from cssdeadwood.css_extract import extract_css_selectors


class CssExtractTest(unittest.TestCase):

    def setUp(self):
        self.cssfile = tempfile.NamedTemporaryFile(mode='w', suffix='.css', delete=False)

    def tearDown(self):
        os.unlink(self.cssfile.name)

    def assertSelectorExtraction(self, css_data, selectors):
        self.cssfile.write(css_data)
        self.cssfile.close()
        extracted = extract_css_selectors(self.cssfile.name)
        self.assertEqual(extracted, selectors)


    def test_simple001(self):
        self.assertSelectorExtraction('''
            p { color: red; }
            div#content h1 { font-size: 10pt;}
            ''',
            set(['p', 'div#content h1'])
        )

    def test_simple002(self):
        self.assertSelectorExtraction('''
            p.red { color: red; }
            div#content h1 { font-size: 10pt;}
            ''',
            set(['p.red', 'div#content h1'])
        )

    def test_simple003(self):
        self.assertSelectorExtraction('''
            p { color: red; }
            div.content h1 { font-size: 10pt;}
            ''',
            set(['p', 'div.content h1'])
        )

    def test_asterisk001(self):
        self.assertSelectorExtraction('''
            * { color: red; }
            h1 { font-size: 10pt;}
            ''',
            set(['*', 'h1'])
        )

    def test_asterisk002(self):
        self.assertSelectorExtraction('''
            *  p{ color: red; }
            h1 { font-size: 10pt;}
            ''',
            set(['* p', 'h1'])
        )

    def test_comma001(self):
        self.assertSelectorExtraction('''
            p, div.red { color: red; }
            div#content h1 { font-size: 10pt;}
            ''',
            set(['p', 'div.red', 'div#content h1'])
        )

    def test_whitespace001(self):
        self.assertSelectorExtraction('''
            p   \n      a, 
                div   \t span.red { color: red; }
            div#content
              h1 { font-size: 10pt;}
            ''',
            set(['p a', 'div span.red', 'div#content h1'])
        )

    def test_whitespace001(self):
        self.assertSelectorExtraction('''
            p>a,div span.red{color:red;}div#content h1{font-size:10pt;}.yellow{color:yellow}
            ''',
            set(['p>a', 'div span.red', 'div#content h1', '.yellow'])
        )

    def test_newlines001(self):
        self.assertSelectorExtraction('''
            p, 
            div.red
            { color: red; }
            div#content 
            h1
            { font-size: 10pt;}
            ''',
            set(['p', 'div.red', 'div#content h1'])
        )

    def test_comments001(self):
        self.assertSelectorExtraction('''
            p   /* span */     a,  div    span.red { color: red; }
            ''',
            set(['p a', 'div span.red'])
        )
    def test_comments002(self):
        self.assertSelectorExtraction('''
            span.red { color: red; }
            /* 
            span.blue { color: blue }
            */
            span.green { color: green }
            ''',
            set(['span.red', 'span.green'])
        )


    def test_media001(self):
        self.assertSelectorExtraction('''
            p, div.red { color: red; }
            @media screen {
                div#content h1 { font-size: 10pt;}
            }
            ''',
            set(['p', 'div.red', 'div#content h1'])
        )

    def test_media002(self):
        self.assertSelectorExtraction('''
            p, div.red { color: red; }
            @media screen {
                div#content h1 { font-size: 10pt;}
                span.blue { color: blue; }
            }
            ''',
            set(['p', 'div.red', 'div#content h1', 'span.blue'])
        )

    def test_media003(self):
        self.assertSelectorExtraction('''
            p, div.red { color: red; }
            @media screen and (device-width:768px) {
                div#content h1 { font-size: 10pt;}
                span.blue { color: blue; }
            }
            ''',
            set(['p', 'div.red', 'div#content h1', 'span.blue'])
        )

    def test_media004(self):
        self.assertSelectorExtraction('''
            p, div.red { color: red; }
            @media screen {
                div#content h1 { font-size: 10pt;}
                span.blue { color: blue; }
            }
            ''',
            set(['p', 'div.red', 'div#content h1', 'span.blue'])
        )


